from __future__ import with_statement

from Foundation import *
from AppKit import *
import objc

# In Mac OS version 10.4 and earlier (and also possible 10.5), the text rendering system will behave
# badly if the affine transform of the graphics context involves a large scale factor. That is, a
# font of size 2 scaled by 10x will look different from a font of size 20 at no scale factor. This
# module exists to work around that problem, by adding an extra affine transform to scale back from
# the circuit view's scale to normal scale so that text can be drawn normally.

rescaler = NSAffineTransform.transform()
rescaler.scaleBy_(1/16.0)

def draw_text_centered(string, loc, attribs):
	"""Draws the given string centered at 'loc' with attributes 'attribs'. Assumes that the current
coordinate system is a circuit coordinate system."""
	size = NSString.sizeWithAttributes_(str(string), attribs)
	size = (size[0]/16.0, size[1]/16.0)
	draw_text(string, (loc[0]-size[0]/2, loc[1]-size[1]/2), attribs)

def draw_text(string, loc, attribs):
	"""Draws a given string with its lower-left corner at 'loc' with attributes 'attribs'. Assumes
that the current coordinate system is a circuit coordinate system."""
	with in_affine_transform(rescaler):
		NSString.drawAtPoint_withAttributes_(str(string), (loc[0]*16, loc[1]*16), attribs)

class in_affine_transform(object):
	"""in_affine_transform exists for use with Python's "with" statement. The intended usage is:

	with in_affine_transform(some_transform):
		...
	
	"""
	def __init__(self, transform):
		self.transform = transform
	
	def __enter__(self):
		NSGraphicsContext.currentContext().saveGraphicsState()
		self.transform.concat()
	
	def __exit__(self, x, y, z):
		NSGraphicsContext.currentContext().restoreGraphicsState()

def load_font(name, size):
	"""load_font is because [NSFont fontWithName:size:] returns nil rather than raising an error
if it cannot find the font."""
	font = NSFont.fontWithName_size_(name, size)
	if not font:
		NSLog("Cannot load font: \""+name+"\".")
		return NSFont.fontWithName_size_("Helvetica", size)
	else:
		return font
