from __future__ import division

from Foundation import NSAffineTransform
import math

# This module contains assorted functions for working with points, lines and rectangles. Points are
# represented as tuples of two numbers. Rectangles are represented as tuples of ((x,y),(w,h)) where
# (x,y) is the lower-left corner of the rectangle.

def add_points(p1, p2):
	"""Returns the sum of two points."""
	return (p1[0]+p2[0], p1[1]+p2[1])

def disect_affine_transform(transform):
	"""Given an NSAffineTransform object, reverse-engineers it to retrieve the original translation,
rotation, and flip operations. Assumes that the affine transformation does not rotate by an angle
that is not a multiple of 90 degrees. """
	center = transform_point(transform, (0,0))
	right = transform_point(transform, (1,0))
	up = transform_point(transform, (0,1))
	
	def similar(p1, p2):
		return abs(p1[0]-p2[0])<0.001 and abs(p1[1]-p2[1])<0.001
	
	if similar(right, add_points(center, (1,0))):
		rotation = 0
		flip = similar(up, add_points(center, (0,-1)))
	elif similar(right, add_points(center, (0,1))):
		rotation = 90
		flip = similar(up, add_points(center, (1,0)))
	elif similar(right, add_points(center, (-1,0))):
		rotation = 180
		flip = similar(up, add_points(center, (0,1)))
	elif similar(right, add_points(center, (0, -1))):
		rotation = 270
		flip = similar(up, add_points(center, (-1,0)))
	else:
		raise ValueError, "Invalid transformation."
	
	center = (round(center[0]*2)/2, round(center[1]*2)/2)
	
	return center, rotation, flip

def expand_rect(rect, margin):
	"""Given a rectangle and a margin, return a new rectangle that is larger than the original by
the given margin. """
	return ((rect[0][0]-margin, rect[0][1]-margin), (rect[1][0]+margin*2, rect[1][1]+margin*2))

def iter_lattice_points(start, end):
	"""Given starting and ending points for a line segment, iterates over all of the lattice points 
that the line segment passes through. Assumes that the starting and ending points are lattice 
points, but does not return the starting and ending points themselves."""
	
	start = round_point(start)
	end = round_point(end)
	dx = end[0] - start[0]
	dy = end[1] - start[1]
	
	if dx==0:
		if start[1]>end[1]:
			start,end = end,start
		for y in xrange(start[1]+1, end[1]):
			yield (start[0], y)
			
	elif dy==0:
		if start[0]>end[0]:
			start,end = end,start
		for x in xrange(start[0]+1, end[0]):
			yield (x, start[1])
			
	else:
		# Eliminate common factors from dx and dy to find the shortest step between lattice
		# points
		for f in xrange(2, max(abs(dx),abs(dy))+1):
			while dx%f==0 and dy%f==0:
				dx/=f
				dy/=f
		
		# Convert dx and dy back to integers after division
		dx = int(dx)
		dy = int(dy)
		
		x,y = start
		x+=dx
		y+=dy
		while (x,y)!=end:
			yield (x,y)
			x+=dx
			y+=dy

def line_segments_overlap((a1,a2), (b1,b2)):
	"""Given the starting and ending points of two line segments, checks whether any part of the
line segments overlap each other. Crossing is not considered 'overlapping', and it is not considered
overlapping for the line segments to share an end point."""
	if a1[0]==a2[0]:
		# Line A is vertical
		x = a1[0]
		if b1[0]!=x or b2[0]!=x:
			return False
		if min(b1[1], b2[1]) >= max(a1[1], a2[1]):
			return False
		if max(b1[1], b2[1]) <= min(a1[1], a2[1]):
			return False
		return True
	
	elif a1[1]==a2[1]:
		# Line A is horizontal
		y = a1[1]
		if b1[1]!=y or b2[1]!=y:
			return False
		if min(b1[0], b2[0]) >= max(a1[0], a2[0]):
			return False
		if max(b1[0], b2[0]) <= min(a1[0], a2[0]):
			return False
		return True
	
	else:
		# Line A is diagonal
		# Reorder lines so leftmost end is first
		if a1[0]>a2[0]:
			a1,a2 = a2,a1
		if b1[0]>b2[0]:
			b1,b2 = b2,b1
		
		if b2[0]==b1[0]:
			# lines cannot overlap, because a1[0]!=a2[0]; exit now to avoid a ZeroDivisionError
			return False
		
		a_slope = (a2[1]-a1[1])/(a2[0]-a1[0])
		b_slope = (b2[1]-b1[1])/(b2[0]-b1[0])
		
		if a_slope!=b_slope:
			return False
		
		a_intercept = a1[1] - a_slope*a1[0]
		b_intercept = b1[1] - b_slope*b1[0]
		
		if a_intercept!=b_intercept:
			return False
		
		if not rects_intersect(rect_with_corners(a1,a2), rect_with_corners(b1,b2)):
			return False

		return True

def make_affine_transform(translate, rotate, flip):
	"""Given a vector to translate by, a number of degrees to rotate by, and a Boolean value for
whether or not to flip, returns a NSAffineTransform object that performs the given transformation.
"""
	t = NSAffineTransform.transform()
	
	t.translateXBy_yBy_(translate[0], translate[1])
	t.rotateByDegrees_(rotate)
	if flip: t.scaleXBy_yBy_(1, -1)
	
	return t

def norm_vector(v):
	"""Normalizes a vector such that its length is 1. Returns the normalized vector."""
	d = math.sqrt(v[0]**2 + v[1]**2)
	return (v[0]/d, v[1]/d)

def point_distance(p1, p2):
	"""Returns the distance between two points."""
	return (p1[0]-p2[0])**2 + (p1[1]-p2[1])**2

def point_in_rect(p, r):
	"""Tests whether a point is in a rectangle. If the point is on the boundary of the rectangle,
returns False."""
	if p[0] <= r[0][0]:
		return False
	if p[1] <= r[0][1]:
		return False
	if p[0] >= r[0][0]+r[1][0]:
		return False
	if p[1] >= r[0][1]+r[1][1]:
		return False
	return True

def round_point(p):
	"""Rounds the X and Y coordinates of a point to the nearest integer, and returns the result."""
	return (int(round(p[0])), int(round(p[1])))

def rect_around_point(p, r):
	"""Returns a rect centered on a given point such that its left, right, upper, and lower
boundaries are a constant distance from the point."""
	return ((p[0]-r, p[1]-r), (2*r, 2*r))

def rect_bounding_objects(objects):
	"""Given an iterator of points and rectangles, rect_bounding_objects creates a rectangle that
bounds all of the points and rectangles and returns it. If there are no objects in the iterator,
returns None."""
	left = float("inf")
	right = float("-inf")
	bottom = float("inf")
	top = float("-inf")
	
	for obj in objects:
		if isinstance(obj[0],(float,int)):
		   # It's a point
		   left = min(obj[0], left)
		   right = max(obj[0], right)
		   bottom = min(obj[1], bottom)
		   top = max(obj[1], top)
		else:
			# It's a rect
			left = min(obj[0][0], left)
			right = max(obj[0][0], right)
			bottom = min(obj[0][1], bottom)
			top = max(obj[0][1], top)
			
			othercorner = add_points(obj[0], obj[1])
			left = min(othercorner[0], left)
			right = max(othercorner[0], right)
			bottom = min(othercorner[1], bottom)
			top = max(othercorner[1], top)
	
	if left==float("inf"):
		return None
	else:
		return ((left, bottom), (right-left, top-bottom))

def rect_center(r):
	"""Returns the center of the given rectangle."""
	return (r[0][0]+r[1][0]/2, r[0][1]+r[1][1]/2) 

def rects_intersect(r1, r2):
	"""Tests whether two rectangles intersect. If the rectangles share only a boundary, they are not
considered to intersect."""
	if r1[0][0] >= r2[0][0]+r2[1][0]:
		return False
	if r1[0][1] >= r2[0][1]+r2[1][1]:
		return False
	if r2[0][0] >=  r1[0][0]+r1[1][0]:
		return False
	if r2[0][1] >=  r1[0][1]+r1[1][1]:
		return False
	return True

def rect_with_corners(a, b):
	"""Given two points, returns a rectangle such that two of its corners are the points."""
	left = min(a[0], b[0])
	right = max(a[0], b[0])
	bottom = min(a[1], b[1])
	top = max(a[1], b[1])
	return ((left,bottom), (right-left,top-bottom))

def rotate_vector(p, degrees):
	"""Rotates a vector by a given number of degrees counter-clockwise. Returns the rotated 
vector."""
	radians = degrees/180.0*math.pi
	s = int(round(math.sin(radians)))
	c = int(round(math.cos(radians)))
	return (p[0]*c+p[1]*s, -p[0]*s+p[1]*c)

def subtract_points(p1, p2):
	"""Returns the difference between two points."""
	return (p1[0]-p2[0], p1[1]-p2[1])

def scale_point(p, f):
	"""Returns the given point with both coordinates multiplied by the given scalar factor."""
	return (p[0]*f, p[1]*f)

def scale_rect(r, f):
	"""Returns the given rectangle with all coordinates multiplied by the given scalar factor."""
	return (scale_point(r[0],f), scale_point(r[1],f))

def transform_point(transform, point):
	"""Transforms a point by a given affine transformation."""
	x,y = transform.transformPoint_(point)
	return (x,y)

def transform_rect(transform, rect):
	"""Transforms a rectangle by a given affine transformation. The behavior of this function is
undefined if the transformation rotates by an angle that is not a multiple of 90 degrees."""
	lower_left = transform_point(transform, rect[0])
	upper_right = transform_point(transform, add_points(rect[0],rect[1]))
	return rect_with_corners(lower_left, upper_right)

def untransform_point(transform, point):
	"""Reverses the transformation of the given point. The given transformation must be an
invertible matrix."""
	m11, m12, m21, m22, tX, tY = transform.transformStruct()
	
	# The following code transforms the point:
	#
	# point = (point[0]*m11+point[1]*m21, point[0]*m12+point[1]*m22)
	# point = (point[0]+tX, point[1]+tY)
	#
	# Reversing this code will retrieve the original point.
	
	point = (point[0]-tX, point[1]-tY)
	determinant = m11*m22 - m12*m21
	assert determinant!=0
	point = (point[0]*m22-point[1]*m21, -point[0]*m12+point[1]*m11)
	point = (point[0]/determinant, point[1]/determinant)
	return point

def untransform_rect(transform, rect):
	"""Reverses the transformation of the given rectangle. The given transformation must be an
invertible matrix. The behavior of this function is undefined if the transformation rotates by an
angle that is not a multiple of 90 degrees."""
	p1 = untransform_point(transform, rect[0])
	p2 = untransform_point(transform, add_points(rect[0], rect[1]))
	return rect_with_corners(p1,p2)

def valid_point(p, circuit):
	"""Checks if a point is within the boundaries of a given circuit. Returns True if the point is
within the boundaries."""
	if p[0]<0: return False
	if p[1]<0: return False
	if p[0]>=circuit.width: return False
	if p[1]>=circuit.height: return False
	return True
