from Foundation import *
from AppKit import *
import objc

from PyObjCTools import NibClassBuilder

import time
from solver4 import SolverError

# Local imports from within Circuit Editor
from CE.Circuit import Circuit
from CE.Component import Component
from CE.Wire import Wire
from circuit_inspector import CircuitInspector
from circuit_view import group_bounds
import circuit_serialize
import alerts
from notifications import *
from geometry import *


# If True, then debug information will be dumped
enable_debug = False


error_message_translation = {
	"Iterated too many times trying to find root of single constraint.":
		"Iterated too many times trying to find root of single constraint. This " \
		"problem was probably caused by a short-circuit somewhere in your circuit.",
	"Iterated too many times trying to find solution to system.":
		"Iterated too many times trying to find solution to system. This problem " \
		"might have been caused by a short-circuit somewhere in your circuit."
	}


class CircuitDocument(NibClassBuilder.AutoBaseClass):

	# The following outlets are defined in the nib:
	# 'view'
	
	def init(self):
		self = super(CircuitDocument, self).init()
		
		# The path from which the circuit should be loaded once the window has finished loading
		self.path = None
		
		# The circuit that is currently being edited
		self.circuit = None
		
		# The speed of the simulation as a multiplier of real time (e.g. "2" means twice as fast as
		# real time)
		self.simulation_speed = 0.0001
		
		# The number of times per second that the simulation updates
		self.simulation_update_rate = 10.0
		
		return self
	
	def dealloc(self):
		# NSNotificationCenter does not retain observer objects, so we must unregister ourselves
		# manually to avoid a problem
		unobserve_notification(self)
		
		super(CircuitDocument, self).dealloc()
	
	# Methods related to duties as a NSDocument
	
	def makeWindowControllers(self):
		main = NSWindowController.alloc().initWithWindowNibName_owner_("CircuitDocument", self)
		main.setShouldCloseDocument_(True)
		self.addWindowController_(main)
	
	def readFromFile_ofType_(self, path, typ):
		if self.view is None:
			# We've not yet fully loaded the window, so we just remember the path to load from later
			self.path = path
		else:
			# We're reverting from a file
			objs = circuit_serialize.deserialize_elements(file(path).read())
			self.setCircuit_(Circuit.alloc().initWithElements_(objs))
		return True
	
	def writeToFile_ofType_(self, path, typ):
		objs = self.circuit.wires | self.circuit.components
		file(path, "w").write(circuit_serialize.serialize_elements(objs))
		return True
	
	def windowControllerDidLoadNib_(self, controller):
		if self.path is None:
			self.setCircuit_(Circuit.alloc().initEmpty())
		else:
			objs = circuit_serialize.deserialize_elements(file(self.path).read())
			self.setCircuit_(Circuit.alloc().initWithElements_(objs))
		del self.path
	
	def close(self):
		# Cancel the running timer
		if self.circuit and self.simulation_enabled and self.simulation_updating:
			self.update_timer.invalidate()
		
		super(CircuitDocument, self).close()
	
	# General-purpose methods
	
	def setCircuit_(self, circuit):
		if self.circuit:
		
			if enable_debug:
				self.circuit.circuit.set_debug_file(None)
				self.debug_file.close()
			
			self.circuit.setUndoFunc_(None)
			unobserve_notification(self, None, self.circuit)
			
			self.setSimulationEnabled_(False)
			del self.simulation_enabled
		
		self.circuit = circuit
		self.view.setCircuit_(circuit)
		
		if self.circuit:
			
			if enable_debug:
				# Open a file for the circuit to write debugging information to
				timestring = time.strftime("circuit-%b%d-%H%M%S")
				path = "/Users/Tim/src/solver4debugger/samples/%s.txt" % timestring
				self.debug_file = file(path, "w")
			
				# Write the ordering of the circuit's elements to that debug file to make it easier 
				# to debug problems that only come up with certain element orders. The argument of 
				# 'True' to serialize_elements() indicates that the elements should be written as a 
				# list rather than a set.
				self.debug_file.write(
					circuit_serialize.serialize_elements(circuit.wires|circuit.components, True))
				self.debug_file.write("\n\n")
				
				self.circuit.circuit.set_debug_file(self.debug_file)
			
			self.circuit.setUndoFunc_(self.registerUndo_)
			
			observe_notification(self, "CircuitElementAdded",
					self.circuitChangedCB_, self.circuit)
			observe_notification(self, "CircuitElementRemoved",
					self.circuitChangedCB_, self.circuit)
			observe_notification(self, "CircuitElementRepositioned",
					self.circuitChangedCB_, self.circuit)
			observe_notification(self, "CircuitElementPropertyChange",
					self.circuitChangedCB_, self.circuit)
			observe_notification(self, "CircuitElementManipulated",
					self.circuitChangedCB_, self.circuit)
			
			self.simulation_enabled = False
			self.setSimulationEnabled_(True)
	
	@objc.signature("v@:@")
	def circuitChangedCB_(self, note):
		self.setNeedsRecalculate()
	
	def setSimulationEnabled_(self, enabled):
		# If the simulation is 'enabled', that means that the simulator will calculate voltages and
		# currents at each point in the circuit, but time will not advance unless the simulation is
		# also 'updating'.
		
		assert self.circuit
		
		if enabled and not self.simulation_enabled:
			self.simulation_enabled = True
			self.recalculate_pending = False
			self.setNeedsRecalculate()
			self.simulation_updating = False
					
		if not enabled and self.simulation_enabled:
			self.setSimulationUpdating_(False)
			del self.recalculate_pending
			del self.simulation_updating
			self.simulation_enabled = False
			
			# Make meter screens go dark
			self.circuit.calculation_valid = False
			self.view.setNeedsDisplay_(True)
	
	def setSimulationUpdating_(self, updating):
		# If the simulation is 'updating', that means that the simulation is 'enabled', but also
		# that capacitors, inductors, and other time-dependent components are changing.
		assert self.circuit
		assert self.simulation_enabled
		
		if updating and not self.simulation_updating:
			self.simulation_updating = True
			self.update_timer = \
				NSTimer.scheduledTimerWithTimeInterval_target_selector_userInfo_repeats_(
				1.0/self.simulation_update_rate, self, self.updateTimerCB_, None, True)
		
		if not updating and self.simulation_updating:
			self.update_timer.invalidate()
			del self.update_timer
			self.simulation_updating = False
	
	def updateOneSecond(self):
		# updateOneSecond schedules the simulation to be advanced by one second of real time, which 
		# is the same thing as 'self.simulation_speed' seconds of simulation-time.
		NSRunLoop.currentRunLoop().performSelector_target_argument_order_modes_(
					self.updateCB_, self, None, 0, [NSDefaultRunLoopMode])
	
	@objc.signature("v@:@")
	def updateTimerCB_(self, timer):
		# _update_timer_cb is called 'simulation_update_rate' times per second.
		if self.simulation_enabled: # In case timer fires after simulation has been disabled
			self.performUpdate_(self.simulation_speed / self.simulation_update_rate)
	
	@objc.signature("v@:@")
	def updateCB_(self, userinfo):
		# Update simulation by one second of real time.
		if self.simulation_enabled:
			self.performUpdate_(self.simulation_speed)
		
	def performUpdate_(self, time):
		# Updates the simulation by 'time' seconds of simulation-time.
		try:
			self.circuit.update_(time)
		except SolverError, e:
			msg = error_message_translation.get(str(e), str(e))
			alerts.show_alert(self.view.window(), "Simulation update failed", msg)
			self.setSimulationEnabled_(False)
	
	def setNeedsRecalculate(self):
		# Schedules the simulation to be recalculated, not immediately but in the near future
		if self.simulation_enabled:
			if not self.recalculate_pending:
				NSRunLoop.currentRunLoop().performSelector_target_argument_order_modes_(
					self.recalculateCB_, self, None, 0, [NSDefaultRunLoopMode])
				self.recalculate_pending = True
	
	@objc.signature("v@:@")
	def recalculateCB_(self, userinfo):
		if self.simulation_enabled: # In case timer fires after simulation has been disabled
			try:
				self.circuit.recalculate()
			except SolverError, e:
				# Convert the error message from the solver into something more user-friendly if
				# possible
				msg = error_message_translation.get(str(e), str(e))
				msg2 = "When you have corrected the problem that caused this error, choose " \
					"\"Enable Simulation\" from the \"Simulation\" menu to test the circuit again."
				alerts.show_alert(self.view.window(), "Simulation failed", msg+"\n\n"+msg2)
				self.setSimulationEnabled_(False)
			else:
				self.recalculate_pending = False
	
	# Actions
	
	def validateMenuItem_(self, menuItem):
		
		# menuItem.action() returns the selector of the menu item as a string, such as "delete:".
		action = menuItem.action()
		
		if action=="toggleSimulationEnabled:":
			if self.circuit:
				if self.simulation_enabled:
					menuItem.setTitle_("Disable Simulation")
				else:
					menuItem.setTitle_("Enable Simulation")
				return True
			else:
				return False
		
		if action=="toggleSimulationUpdating:":
			if self.circuit and self.simulation_enabled:
				if self.simulation_updating:
					menuItem.setTitle_("Stop Simulation")
				else:
					menuItem.setTitle_("Start Simulation")
				return True
			else:
				return False
		
		if action=="updateSimulation:":
			if self.circuit and self.simulation_enabled and not self.simulation_updating:
				return True
			else:
				return False
		
		if action == "paste:":
			# "Paste" should be disabled if the clipboard does not contain valid data, or if there
			# is already something floating
			pb = NSPasteboard.generalPasteboard()
			return ("CircuitElements" in pb.types()) and not CircuitFloatManager.shared.floating
		
		return super(CircuitDocument, self).validateMenuItem_(menuItem)
	
	@objc.IBAction
	def toggleSimulationEnabled_(self, sender):
		if self.circuit:
			self.setSimulationEnabled_(not self.simulation_enabled)
	
	@objc.IBAction
	def toggleSimulationUpdating_(self, sender):
		if self.circuit and self.simulation_enabled:
			self.setSimulationUpdating_(not self.simulation_updating)
	
	@objc.IBAction
	def updateSimulation_(self, sender):
		if self.circuit and self.simulation_enabled:
			self.updateOneSecond()
	
	# Undo/redo
	
	def registerUndo_(self, func):
		# Registers a given function with the undo manager. The function should take no arguments.
		um = self.undoManager()
		um.registerUndoWithTarget_selector_object_(self, self.performUndoCB_, func)
	
	@objc.signature("v@:@")
	def performUndoCB_(self, func):
		# This method exists because Cocoa wants its undo callback to take the form of a selector, 
		# but in Python it is more convenient to use a callable object instead. So,
		# CircuitDocument.registerUndo_ takes a callable object and stores it as the userdata of the
		# undo callback, so that this method can call it.
		func()
