import objc
from Foundation import *
from AppKit import *

from geometry import *
from catch_exceptions import catch_exceptions

import circuit_photograph
from CE.Wire import Wire
from CE.utils import *


class DependentDragSession(NSObject):
	"""DependentDragSession is used when a group of objects is being moved and at least one wire has
	one end in the selection and the other end out of the selection."""

	def initWithView_parts_(self, view, parts):
		self.view = view
		self.circuit = self.view.circuit
		self.parts = parts
		self.begun = False
		self.finished = False
		return self
	
	def run(self):
		# Create a picture representing the dragged objects
		bounds = ((0,0),(self.view.circuit.width,self.view.circuit.height))
		bounds = expand_rect(bounds, 1)
		
		objs_to_photograph = set()
		for part in self.parts:
			if isinstance(part, Component):
				objs_to_photograph.add(part)
			elif part.other() in self.parts:
				objs_to_photograph.add(part.wire)
		
		extra_dots = [part.location
			for part in self.parts
			if isinstance(part, Wire) and part.other() not in self.parts]
		
		rep = circuit_photograph.photograph(bounds, objs_to_photograph, extra_dots)
		image = NSImage.alloc().initWithSize_(scale_point(bounds[1], 16.0))
		image.addRepresentation_(rep)
		
		# Prepare the pasteboard
		pboard = NSPasteboard.pasteboardWithName_(NSDragPboard)
		pboard.declareTypes_owner_(["DependentCircuitComponents"], None)
		# No data goes on the pasteboard because the only valid drag target will be ourselves,
		# so we will just transfer the objects directly
		pboard.setString_forType_("", "DependentCircuitComponents")
		
		# This call blocks until the drag is complete
		self.view.dragImage_at_offset_event_pasteboard_source_slideBack_(
			image,
			(-1.0, -1.0),
			NSZeroSize, # Ignored
			NSApp().currentEvent(),
			pboard,
			self,
			True)
	
	def draw(self):
		if self.begun and not self.finished:
			for obj in objs_of_parts(self.parts):
				if obj in self.failures: color = NSColor.redColor()
				else: color = NSColor.greenColor()
				obj.draw_outline(transform=(self.transform, self.parts), color=color)
	
	# Drag source role
	
	def draggingSourceOperationMaskForLocal_(self, local):
		if local: return NSDragOperationMove
		else: return NSDragOperationNone
	
	def ignoreModifierKeysWhileDragging(self):
		# Copying is not allowed during a dependent drag session, so modifier keys are not
		# necessary. Furthermore, if we don't ignore modifier keys, then when the user Command-
		# clicks to create a wire Cocoa will interpret it as a modifier key and cause the drag
		# operation to be NSDragOperationGeneric.
		return True
	
	# It would be nice to draw the bodies of wires that are half-in and half-out of the selection,
	# and this would be the place to do it, but it would be difficult.
	
	def draggedImage_beganAt_(self, image, point):
		# TODO: begin drawing wires from board to dragged group
		pass
	
	def draggedImage_movedTo_(self, image, point):
		# TODO: draw wires from board to dragged group
		pass
	
	@catch_exceptions
	def draggedImage_endedAt_operation_(self, image, point, op):
		# TODO: end drawing wires from board to dragged group
		
		# There is nothing to do because CircuitView's performDrag_ method has performed the
		# necessary work.
		pass
	
	# Drag destination role (calls are forwarded from CircuitView)
	
	def dragging_entered(self, sender):
	
		self.update(sender)
		
		if self.failures: return NSDragOperationNone
		else: return NSDragOperationMove
	
	def dragging_updated(self, sender):
		
		# Re-draw over old location of parts
		bounds = group_bounds(parts_of_objs(objs_of_parts(self.parts)),
			transform=(self.transform, self.parts))
		bounds = expand_rect(bounds, 1.0)
		self.view.setNeedsDisplayInRect_(bounds)
		
		self.update(sender)
		
		if self.failures: return NSDragOperationNone
		else: return NSDragOperationMove
		
	def dragging_exited(self, sender):
		self.view.setNeedsDisplay_(True)
	
	def prepare_for_drag_operation(self, sender):
		return (self.failures == set())
	
	def perform_drag(self, sender):
		if not self.is_trivial:
			for obj in objs_of_parts(self.parts):
				if isinstance(obj, Component):
					obj.reposition(obj.transformed_position(self.transform))
				else:
					obj.reposition(obj.transformed_location(self.transform, self.parts))
		self.finished = True
		return True
	
	def conclude_drag(self, sender):
		self.view.setNeedsDisplay_(True)
	
	# Compute which components can and cannot fit
	
	def update(self, sender):
		
		self.begun = True
		
		# Compute new proposed drop
		point_in_view = self.view.convertPoint_fromView_(sender.draggedImageLocation(), None)
		delta = add_points(round_point(point_in_view), (1,1))
		self.is_trivial = (delta == (0,0))
		self.transform = make_affine_transform(delta, 0, False)
		
		# See if new proposed drop will work
		self.failures = test_transform(
			(self.circuit.width, self.circuit.height),
			self.circuit.wires | self.circuit.components,
			(self.transform, self.parts))
		
		# Re-draw over new location of parts
		bounds = group_bounds(parts_of_objs(objs_of_parts(self.parts)),
			transform=(self.transform, self.parts))
		bounds = expand_rect(bounds, 1.0)
		self.view.setNeedsDisplayInRect_(bounds)