import objc
from Foundation import *
from AppKit import *

from geometry import *
from catch_exceptions import catch_exceptions

import circuit_photograph
import circuit_serialize
from CE.utils import *

class FreeDragSession(NSObject):
	"""FreeDragSession is used when the user is moving a group of objects and every object is either
	entirely in the selection or entirely out of it (there are no wires where one end is in the
	selection and the other end isn't). When dragging between two different CircuitViews, two
	FreeDragSessions are created: one for the source view and one for the destination view."""

	def initWithView_(self, view):
		self.view = view
		self.circuit = self.view.circuit
		self.in_this_view = False
		self.finished = False
		return self
	
	@classmethod
	def new_as_source(cls, view, objs):
		new = cls.alloc().initWithView_(view)
		new.objs = objs
		new.type = "source"
		return new
	
	@classmethod
	def new_as_dest(cls, view, sender):
		new = cls.alloc().initWithView_(view)
		pboard = sender.draggingPasteboard()
		pboard.availableTypeFromArray_(["CircuitComponents"])
		serialized = pboard.stringForType_("CircuitComponents")
		new.objs = set(circuit_serialize.deserialize_elements(serialized))
		new.type = "dest"
		return new
	
	def run(self):
	
		assert self.type == "source"
	
		# Create a picture representing the dragged objects
		bounds = ((0,0),(self.view.circuit.width,self.view.circuit.height))
		bounds = expand_rect(bounds, 1)
		
		rep = circuit_photograph.photograph(bounds, self.objs, [])
		image = NSImage.alloc().initWithSize_(scale_point(bounds[1], 16.0))
		image.addRepresentation_(rep)
		
		# Prepare the pasteboard
		pboard = NSPasteboard.pasteboardWithName_(NSDragPboard)
		pboard.declareTypes_owner_(["CircuitComponents"], None)
		serialized = circuit_serialize.serialize_elements(self.objs)
		pboard.setString_forType_(serialized, "CircuitComponents")
				
		# This call blocks until the drag is complete
		self.view.dragImage_at_offset_event_pasteboard_source_slideBack_(
			image,
			(-1.0, -1.0),
			NSZeroSize, # Ignored
			NSApp().currentEvent(),
			pboard,
			self,
			True)
	
	def draw(self):
		if self.in_this_view and not self.finished:
			for obj in self.objs:
				if obj in self.failures: color = NSColor.redColor()
				else: color = NSColor.greenColor()
				obj.draw_outline(transform=(self.transform, parts_of_objs(self.objs)), color=color)
	
	# Dragging source role
	
	def draggingSourceOperationMaskForLocal_(self, local):
		return NSDragOperationMove | NSDragOperationCopy | NSDragOperationDelete
	
	def ignoreModifierKeysWhileDragging(self):
		return False
	
	def draggedImage_beganAt_(self, image, point):
		pass
	
	def draggedImage_movedTo_(self, image, point):
		pass
	
	@catch_exceptions
	def draggedImage_endedAt_operation_(self, image, point, op):
		if op in (NSDragOperationMove, NSDragOperationDelete) and not self.in_this_view:
			# We have moved our components to another location, or we are deleting them. In either
			# case they must be removed from this view.
			for obj in self.objs:
				obj.remove()
	
	# Drag destination role (calls are forwarded from CircuitView)
	
	def dragging_entered(self, sender):
		self.in_this_view = True
		return self.update(sender)
	
	def dragging_updated(self, sender):
		
		# Re-draw over old location of objects
		bounds = group_bounds(self.objs, transform=(self.transform, parts_of_objs(self.objs)))
		bounds = expand_rect(bounds, 1.0)
		self.view.setNeedsDisplayInRect_(bounds)
		
		return self.update(sender)
	
	def dragging_exited(self, sender):
		self.in_this_view = False
		self.view.setNeedsDisplay_(True)
	
	def prepare_for_drag_operation(self, sender):
		return (self.failures == set())
	
	def perform_drag(self, sender):
		if sender.draggingSource() is self:
			if not self.is_trivial:
				for obj in self.objs:
					if isinstance(obj, Component):
						obj.reposition(obj.transformed_position(self.transform))
					else:
						parts = set([obj.end, obj.start])
						obj.reposition(obj.transformed_location(self.transform, parts))
		else:
			for obj in self.objs:
				if isinstance(obj, Component):
					obj.center = transform_point(self.transform, obj.center)
					obj.transform = make_affine_transform(obj.center, obj.rotation, obj.flip)
				else:
					obj.start.loc = transform_point(self.transform, obj.start.loc)
					obj.end.loc = transform_point(self.transform, obj.start.loc)
				obj.install(self.circuit)
		self.view.selection = parts_of_objs(self.objs)
		self.finished = True
		return True
	
	def conclude_drag(self, sender):
		self.view.setNeedsDisplay_(True)
	
	# Compute which components can and cannot fit
	
	def update(self, sender):
	
		# Determine what we are going to do
		ops = sender.draggingSourceOperationMask()
		if ops & NSDragOperationMove: op = NSDragOperationMove
		elif ops & NSDragOperationCopy: op = NSDragOperationCopy
		else: return NSDragOperationNone
		
		# Compute new proposed drop
		point_in_view = self.view.convertPoint_fromView_(sender.draggedImageLocation(), None)
		delta = add_points(round_point(point_in_view), (1,1))
		self.is_trivial = (delta == (0,0)) and sender.draggingSource() is self
		self.transform = make_affine_transform(delta, 0, False)
		
		# See if new proposed drop will work
		if op == NSDragOperationMove and sender.draggingSource() is self:
			others = (self.circuit.wires | self.circuit.components) - self.objs
		else:
			others = self.circuit.wires | self.circuit.components
		
		self.failures = test_transform(
			(self.circuit.width, self.circuit.height),
			others,
			(self.transform, parts_of_objs(self.objs)))
		
		# Re-draw over new location of parts
		bounds = group_bounds(self.objs, transform=(self.transform, parts_of_objs(self.objs)))
		bounds = expand_rect(bounds, 1.0)
		self.view.setNeedsDisplayInRect_(bounds)
		
		if self.failures: return NSDragOperationNone
		else: return op