import objc
from Foundation import *
from AppKit import *

from geometry import *
from catch_exceptions import catch_exceptions

from CE.Wire import Wire
from CE.utils import *
import circuit_photograph

class WireDragSession(NSObject):
	"""WireDragSession is used when the user is creating a new wire."""

	def initWithView_startLoc_(self, view, loc):
		self.view = view
		self.circuit = self.view.circuit
		self.wire = Wire((loc, (loc[0]+1,loc[1])))
		self.begun = False
		self.finished = False
		return self
	
	def run(self):
		
		# Make a picture of a wire end
		bounds = ((-0.5,-0.5), (1.0, 1.0))
		rep = circuit_photograph.photograph(bounds, [], [(0,0)])
		image = NSImage.alloc().initWithSize_((16, 16))
		image.addRepresentation_(rep)
		
		# Prepare the pasteboard
		pboard = NSPasteboard.pasteboardWithName_(NSDragPboard)
		pboard.declareTypes_owner_(["CircuitWireEnd"], None)
		# No data goes on the pasteboard because the only valid drag target will be ourselves
		pboard.setString_forType_("", "CircuitWireEnd")
		
		# Compute the initial location of the picture
		pic_loc = self.view.convertPoint_fromView_(NSApp().currentEvent().locationInWindow(), None)
		pic_loc = add_points(pic_loc, (-0.5, -0.5))
				
		# This call blocks until the drag is complete
		self.view.dragImage_at_offset_event_pasteboard_source_slideBack_(
			image,
			pic_loc,
			NSZeroSize, # Ignored
			NSApp().currentEvent(),
			pboard,
			self,
			True)
	
	def draw(self):
		if self.begun and not self.finished:
			if self.failed: color = NSColor.redColor()
			else: color = NSColor.greenColor()
			self.wire.draw_outline(transform=(self.transform, set([self.wire.end])), color=color)
	
	# Dragging source role
	
	def draggingSourceOperationMaskForLocal_(self, local):
		if local: return NSDragOperationMove
		else: return NSDragOperationNone
	
	def ignoreModifierKeysWhileDragging(self):
		return True
	
	def draggedImage_beganAt_(self, image, point):
		# TODO: begin drawing a wire
		pass
	
	def draggedImage_movedTo_(self, image, point):
		# TODO: draw the wire
		pass
	
	@catch_exceptions
	def draggedImage_endedAt_operation_(self, image, point, op):
		# TODO: end drawing the wire
		pass
	
	# Drag destination role (calls are forwarded from CircuitView)
	
	def dragging_entered(self, sender):
		return self.update(sender)
	
	def dragging_updated(self, sender):
	
		# Re-draw over new location of parts
		bounds = group_bounds([self.wire.start, self.wire.end],
			transform=(self.transform, set([self.wire.end])))
		bounds = expand_rect(bounds, 1.0)
		self.view.setNeedsDisplayInRect_(bounds)
	
		return self.update(sender)
	
	def dragging_exited(self, sender):
		self.view.setNeedsDisplay_(True)
	
	def prepare_for_drag_operation(self, sender):
		return not self.failed
	
	def perform_drag(self, sender):
		self.wire.end.location = round_point(transform_point(self.transform,self.wire.end.location))
		self.wire.install(self.circuit)
		self.finished = True
		return True
	
	def conclude_drag(self, sender):
		self.view.setNeedsDisplay_(True)
	
	# Compute whether the wire is valid
	
	def update(self, sender):
		
		self.begun = True
		
		# Compute new proposed drop
		pic_loc = self.view.convertPoint_fromView_(sender.draggedImageLocation(), None)
		point_loc = add_points(pic_loc, (0.5,0.5))
		delta = subtract_points(round_point(point_loc), self.wire.end.location)
		self.transform = make_affine_transform(delta, 0, False)
		
		# See if new proposed drop will work
		self.failed = bool(test_transform(
			(self.circuit.width, self.circuit.height),
			self.circuit.wires | self.circuit.components | set([self.wire]),
			(self.transform, set([self.wire.end]))
			))
		
		# Re-draw over new location of parts
		bounds = group_bounds([self.wire.start, self.wire.end],
			transform=(self.transform, set([self.wire.end])))
		bounds = expand_rect(bounds, 1.0)
		self.view.setNeedsDisplayInRect_(bounds)
		
		if self.failed: return NSDragOperationNone
		else: return NSDragOperationMove