from Foundation import *
from AppKit import *
import objc
from PyObjCTools import NibClassBuilder

from CE.Component import Component
from notifications import *
from catch_exceptions import catch_exceptions

import traceback

class CircuitInspector(NibClassBuilder.AutoBaseClass):	
	
	# The circuit inspector is a global object that examines circuit components and edits their
	# properties.
	
	# The following outlets are defined in the nib:
	#
	# 'editingView'
	# 'nameLabel'
	
	def init(self):
		self.initWithWindowNibName_("CircuitInspector")
		self.showWindow_(self)
		
		# The current set of objects that the inspector is inspecting
		self.current_contents = None
		
		# If 'self.current_contents' has only one element, then it will be in 'self.current_target'.
		self.current_target = None
		
		self.bind_toObject_withKeyPath_options_(
				"contents",
				NSApp(),
				"mainWindow.windowController.document.view.selection",
				{})
		
		# Watch for property changes from other sources so we can keep the value displayed in sync
		observe_notification(self, "CircuitElementPropertyChange", self.propertyChangedCB_)
		
		# Watch for losing key-window status so we can commit editing when it happens
		observe_notification(self, NSWindowDidResignKeyNotification, self.windowDidResignKeyCB_)
		
		return self
	
	def setContents_(self, contents):
		self.current_contents = contents
				
		if self.current_contents is None:
			c = []
		else:
			c = [x for x in self.current_contents if isinstance(x,Component)]
				
		if self.current_target:		
			# Dismantle the editors and stuff created for the previous target
			self.commitEditing()
			
			for (label, control) in self.editors:
				label.removeFromSuperview()
				control.removeFromSuperview()
			del self.editors
			
		self.current_target = None
		
		if len(c)==0:
			self.nameLabel.setStringValue_("No Selection")
			
		elif len(c)==1:
			self.setCurrentTarget_(c[0])
			
		else:
			self.nameLabel.setStringValue_("Multiple Objects Selected")
				
	def contents(self):
		return self.current_contents
	
	def setCurrentTarget_(self, target):
		self.current_target = target
		
		self.nameLabel.setStringValue_(self.current_target.class_name)
		
		# Build editors for the properties of the new target
		self.editors = []
		control_width = self.editingView.frame()[1][0] - 110
		for i, (name, ui_name, type) in enumerate(self.current_target.editable_properties):
			
			y = self.editingView.frame()[1][1] - 40 - i*30
		
			label = NSTextField.alloc().initWithFrame_(((0,y), (100,21)))
			self.editingView.addSubview_(label)
			
			label.setStringValue_(ui_name)
			label.setEditable_(False)
			label.setBordered_(False)
			label.setBezeled_(False)
			label.setDrawsBackground_(False)
			label.setAlignment_(NSRightTextAlignment)
			
			if type in ("float", "smallfloat", "string"):
				control = NSTextField.alloc().initWithFrame_(((110,y), (control_width,21)))
				self.editingView.addSubview_(control)
				
				control.setEditable_(True)
				control.setBezeled_(True)
				control.cell().setSendsActionOnEndEditing_(True)
				
				control.setTag_(i)
				control.setTarget_(self)
				control.setAction_(self.propertyEdited_)
								
			else:
				raise ValueError, "Don't know how to format type %s" % type
			
			self.editors.append((label,control))
			
			self.updateTargetPropertyNumber_(i)
	
	@objc.IBAction
	def propertyEdited_(self, sender):
		# propertyEdited_ is called whenever the user modifies one of the fields in the inspector
		# panel
		i = sender.tag()
		name, ui_name, type = self.current_target.editable_properties[i]
		label, control = self.editors[i]
		
		if type=="float" or type=="smallfloat":
			value = control.doubleValue()
		elif type=="string":
			value = str(control.stringValue())
		else:
			raise ValueError, "Don't know how to format type %s" % type
		
		self.current_target.change_property(name, value)
		
	@objc.signature("v@:@")
	@catch_exceptions
	def propertyChangedCB_(self, note):
			
		# propertyChangedCB_ is called whenever a property changes from some source other than the
		# inspector panel
		component, property = note.userInfo()
		if component==self.current_target:
			for i, (name, ui_name, type) in enumerate(self.current_target.editable_properties):
				if name==property:
					self.updateTargetPropertyNumber_(i)
					break
			else:
				raise KeyError, "No property named %s" % property
			
	def updateTargetPropertyNumber_(self, number):
		# updateTargetPropertyNumber_ copies the value of the property with the given number from
		# the current target into the inspector panel
		name, ui_name, type = self.current_target.editable_properties[number]
		label, control = self.editors[number]
		value = self.current_target.get_property(name)
		
		if type=="float" or type=="smallfloat":
			control.setDoubleValue_(value)
		elif type=="string":
			control.setStringValue_(value)
		else:
			raise ValueError, "Don't know how to format type %s" % type
	
	@objc.signature("v@:@")
	def windowDidResignKeyCB_(self, note):
		if self.current_target:
			self.commitEditing()
	
	def commitEditing(self):
		for (label, control) in self.editors:
			if control.currentEditor():
				control.cell().endEditing_(control.currentEditor())
