from Foundation import *
from AppKit import *
import objc

from PyObjCTools import NibClassBuilder

import math

# Local imports from within Circuit Editor
from geometry import *
from circuit_drag_free import FreeDragSession
from circuit_drag_dependent import DependentDragSession
from circuit_drag_wire import WireDragSession
from circuit_drag_manipulate import ManipulateDragSession
from circuit_drag_select import SelectDragSession
from notifications import *
import alerts
from catch_exceptions import catch_exceptions

from CE.Wire import Wire, WireEnd
from CE.Component import Component
from CE.utils import *

# CircuitView uses a coordinate system where 1 unit = the distance between two grid points.
# The origin of the coordinate system is the lower-leftmost grid point.
# This way, circuit coordinates are identical to view coordinates.

class CircuitView(NibClassBuilder.AutoBaseClass):

	# The following outlets are defined in the nib:
	# 'document'
	
	# 'selection' is an Objective-C variable so that it can be key-value observed by the
	# CircuitInspector object.
	selection = objc.ivar(u"selection")
	
	def initWithFrame_(self, frame):
		self = super(CircuitView, self).initWithFrame_(frame)
		
		if self:
			# The circuit that is currently being shown and edited
			self.circuit = None
			
			# If True, then each pin in the circuit will be labeled with its current voltage.
			self.do_show_voltage_labels = False
			
			self.registerForDraggedTypes_(
				["CircuitComponents", "DependentCircuitComponents", "CircuitWireEnd"])
		
		return self
	
	def dealloc(self):
		unobserve_notification(self)
		
		super(CircuitView, self).dealloc()
	
	def setCircuit_(self, circuit):
		if self.circuit:
			unobserve_notifications(self, None, self.circuit)
			
			del self.selection
			del self.drag_session
			del self.drag_session_temporary
		
		self.circuit = circuit
		
		if self.circuit:
			# Resize self to proper size to hold entire circuit
			self.setFrameSize_((self.circuit.width*16, self.circuit.height*16))
			
			# Change self's coordinate system so that 1 unit = 16 pixels
			self.setBounds_(((-0.5, -0.5), (self.circuit.width, self.circuit.height)))
			
			observe_notification(self, "CircuitRecalculated",
					self.circuitRecalculatedCB_, self.circuit)
			
			observe_notification(self, "CircuitElementAdded",
					self.circuitElementAddedCB_, self.circuit)
			observe_notification(self, "CircuitElementRepositioned",
					self.circuitElementRepositionedCB_, self.circuit)
			observe_notification(self, "CircuitElementRemoved",
					self.circuitElementRemovedCB_, self.circuit)
			observe_notification(self, "CircuitElementManipulated",
					self.circuitElementChangedCB_, self.circuit)
			observe_notification(self, "CircuitElementPropertyChange",
					self.circuitElementPropertyChangedCB_, self.circuit)
			
			self.selection = set()
			self.drag_session = None
			self.drag_session_temporary = False
				
	def acceptsFirstResponder(self):
		return True
	
	@catch_exceptions
	def drawRect_(self, cliprect):
		rects_being_drawn = self.getRectsBeingDrawn_count_()[0]
		
		# Draw white background
		NSColor.whiteColor().set()
		for rect in rects_being_drawn:
			NSRectFill(rect)
				
		if not self.circuit:
			return
		
		# Adjust default line width to one pixel
		NSBezierPath.setDefaultLineWidth_(1/16.0)
		
		# Draw grid points
		for rect in rects_being_drawn:
			for x in xrange(int(math.ceil(rect[0][0])), int(rect[0][0]+rect[1][0])+1):
				for y in xrange(int(math.ceil(rect[0][1])), int(rect[0][1]+rect[1][1])+1):
					NSColor.grayColor().set()
					r = 0.01
					NSBezierPath.strokeRect_(((x-r,y-r), (r*2,r*2)))
				
		# Draw components
		for component in self.circuit.components:
			if self.needsToDrawRect_(expand_rect(component.bounds(), 0.5)):
				component.draw()				
		
		# Draw wires
		for wire in self.circuit.wires:
			if self.needsToDrawRect_(expand_rect(wire.bounds(), 0.5)):
				wire.draw()
		
		# Draw labels on components
		if self.do_show_voltage_labels:
			for component in self.circuit.components:
				if self.needsToDrawRect_(expand_rect(component.bounds(), 1.0)):
					component.draw_labels()
		
		# Draw highlights on selected objects
		for obj in self.selection:
			obj.draw_selection()
		
		# Draw whatever is necessary for our current drag action
		if self.drag_session:
			self.drag_session.draw()
	
	@catch_exceptions
	def resetCursorRects(self):
		self.discardCursorRects()
		for component in self.circuit.components:
			for rect in component.manipulation_zones:
				rect2 = transform_rect(component.transform, rect)
				self.addCursorRect_cursor_(rect2, NSCursor.pointingHandCursor())
	
	def acceptsFirstMouse_(self, event):
		return True
		
	def applyTransformIfPossible_(self, func):
		# Applies 'func' to the current selection if possible. Otherwise, beeps at the user.
		
		# Remember that testTransform_ returns a set of failures. So if the failure set is empty, 
		# the transformation is valid.
		failures = test_transform(
				(self.circuit.width, self.circuit.height),
				self.circuit.components.union(self.circuit.wires),
				(func, self.selection))
				
		if self.selection and not failures:
			objs = self.selection.copy()
			while objs:
				obj = objs.pop()
				if isinstance(obj, WireEnd):
					new_loc = obj.wire.transformed_location(func,self.selection)
					obj.wire.reposition(new_loc)
					objs.discard(obj.other())
				else:
					new_pos = obj.transformed_position(func)
					obj.reposition(new_pos)
		else:
			NSBeep()
	
	def validateMenuItem_(self, menuItem):
		
		# Returns the selector of the menu item as a string, such as "delete:".
		action = menuItem.action()
		
		if action in ["cut:", "copy:", "delete:", "rotateLeft:", "rotateRight:", "rotate180:", 
						"flipVertical:", "flipHorizontal:"]:
			# These commands operate on the selection, and are meaningless if the selection is
			# empty.
			return (self.selection != set())
		
		if action == "toggleShowVoltageLabels:":
			# 'validateMenuItem:' can also be used to prepare menu item labels. How convenient.
			if self.do_show_voltage_labels: menuItem.setTitle_("Hide Voltage Labels")
			else: menuItem.setTitle_("Show Voltage Labels")
			return True
		
		return True
	
	# "Edit" menu
	
	@objc.IBAction
	def selectAll_(self, sender):
		self.selection = parts_of_objs(self.circuit.components | self.circuit.wires)
	
	@objc.IBAction
	def delete_(self, sender):
		for obj in objs_of_parts(self.selection):
			obj.remove()		
		self.selection = set()
	
	@objc.IBAction
	def flipHorizontal_(self, sender):
		bounds = group_bounds(self.selection)
		if bounds:
			centerx = bounds[0][0] + bounds[1][0] / 2.0
			
			flip = NSAffineTransform.transform()
			flip.translateXBy_yBy_(centerx, 0)
			flip.scaleXBy_yBy_(-1, 1)
			flip.translateXBy_yBy_(-centerx, 0)
			
			self.applyTransformIfPossible_(flip)
		else:
			NSBeep()
	
	@objc.IBAction
	def flipVertical_(self, sender):
		bounds = group_bounds(self.selection)
		if bounds:
			centery = bounds[0][1] + bounds[1][1] / 2.0
			
			flip = NSAffineTransform.transform()
			flip.translateXBy_yBy_(0, centery)
			flip.scaleXBy_yBy_(1, -1)
			flip.translateXBy_yBy_(0, -centery)
			
			self.applyTransformIfPossible_(flip)
		else:
			NSBeep()
	
	@objc.IBAction
	def rotate180_(self, sender):
		# Implement as a flip rather than using rotationSelection_ because it's possible to rotate 
		# 180 degrees around a point where one coordinate is an integer and the other is not, which 
		# the generic rotationSelection_ function does not allow.
		bounds = group_bounds(self.selection)
		if bounds:
			centerx = bounds[0][0] + bounds[1][0] / 2.0
			centery = bounds[0][1] + bounds[1][1] / 2.0
			
			flip = NSAffineTransform.transform()
			flip.translateXBy_yBy_(centerx, centery)
			flip.scaleXBy_yBy_(-1, -1)
			flip.translateXBy_yBy_(-centerx, -centery)
			
			self.applyTransformIfPossible_(flip)
		else:
			NSBeep()
	
	@objc.IBAction
	def rotateLeft_(self, sender):
		self.rotateSelection_(90)
	
	@objc.IBAction
	def rotateRight_(self, sender):
		self.rotateSelection_(-90)
		
	def rotateSelection_(self, degrees):
		# rotateSelection_() exists because so much code is shared between rotateLeft_ and
		# rotateRight_.
		bounds = group_bounds(self.selection)
		
		if bounds:
			centerx = bounds[0][0] + bounds[1][0] / 2.0
			centery = bounds[0][1] + bounds[1][1] / 2.0
						
			# If we rotate around a point that is halfway between two lattice points, some points on 
			# the grid can get transformed into points that are not on the grid. On the other hand, 
			# it's safe to rotate around (n+0.5, m+0.5) if (n, m) is a lattice point. This bit of
			# code transforms coordinates of the forms (n+0.5, m) and (n, m+0.5) into the form
			# (n+0.5, m+0.5). The (centerx % 2) is so that repeated rotations pick the same center
			# as much as possible.
			if centerx==round(centerx) and centery!=round(centery):
				centerx = centerx + 0.5 - (centerx % 2)
			elif centery==round(centery) and centerx!=round(centerx):
				centery = centery + 0.5 - (centery % 2)
			center = (centerx,centery)
			
			assert degrees%90 == 0
			
			rotate = NSAffineTransform.transform()
			rotate.translateXBy_yBy_(centerx, centery)
			rotate.rotateByDegrees_(degrees)
			rotate.translateXBy_yBy_(-centerx, -centery)
			
			self.applyTransformIfPossible_(rotate)
		
		else:
			NSBeep()
	
	# "View" menu
	
	def toggleShowVoltageLabels_(self, sender):
		self.do_show_voltage_labels = not self.do_show_voltage_labels
		self.setNeedsDisplay_(True)
	
	# Responses to notifications
	
	@objc.signature("v@:@")
	def circuitRecalculatedCB_(self, note):
		# Redraw ourselves because the circuit's voltage and current value has been recalculated
		for component in self.circuit.components:
			# If do_show_voltage_labels is off, then we only need to redraw a component if it is a
			# meter with a display. Otherwise, we need to redraw it so that the voltage labels get
			# redrawn.
			if self.do_show_voltage_labels or component.dynamic_appearance:
				self.setNeedsDisplayInRect_(expand_rect(component.bounds(), 1))
	
	@objc.signature("v@:@")
	@catch_exceptions
	def circuitElementAddedCB_(self, note):
	
		# Redraw the object that was added
		obj = note.userInfo()
		self.setNeedsDisplayInRect_(expand_rect(obj.bounds(), 1))
		
		# Install cursor rectangles
		if isinstance(obj, Component) and obj.manipulation_zones:
			# The recommended technique would be "self.window().invalidateCursorRectsForView_(self)"
			# but this does not seem to work.
			self.resetCursorRects()
	
	@objc.signature("v@:@")
	@catch_exceptions
	def circuitElementChangedCB_(self, note):
	
		# Redraw the object that changed
		obj = note.userInfo()
		self.setNeedsDisplayInRect_(expand_rect(obj.bounds(), 1))
	
	@objc.signature("v@:@")
	@catch_exceptions
	def circuitElementPropertyChangedCB_(self, note):
	
		# Redraw the object that changed
		(obj, _) = note.userInfo()
		self.setNeedsDisplayInRect_(expand_rect(obj.bounds(), 1))
	
	@objc.signature("v@:@")
	def circuitElementRepositionedCB_(self, note):
	
		# Redraw ourselves because a circuit element moved. We should redraw the entire view
		# because we don't know where the circuit element was before the change.
		self.setNeedsDisplay_(True)
		
		# Redo cursor rectangles
		obj = note.userInfo()
		if isinstance(obj, Component) and obj.manipulation_zones:
			# The recommended technique would be "self.window().invalidateCursorRectsForView_(self)"
			# but this does not seem to work.
			self.resetCursorRects()
	
	@objc.signature("v@:@")
	@catch_exceptions
	def circuitElementRemovedCB_(self, note):
	
		# If the deleted object is part of our current selection, then remove it from the 
		# selection.
		obj = note.userInfo()
		if isinstance(obj, Wire):
			sel = self.selection - set([obj.start, obj.end])
		else:
			sel = self.selection - set([obj])
		self.selection = sel
		
		# Redraw
		self.setNeedsDisplay_(True)
		
		# Remove cursor rectangles
		if isinstance(obj, Component) and obj.manipulation_zones:
			# The recommended technique would be "self.window().invalidateCursorRectsForView_(self)"
			# but this does not seem to work.
			self.resetCursorRects()
	
	@catch_exceptions
	def mouseDown_(self, event):
	
		# The user can use the mouse to interact with the view in a variety of complicated ways.
		# There are at least four different types of drag actions (depending on how you count
		# them), so the code to handle mouse movements is split across multiple files.
		#
		# When mouseDown: is called, the view determines what kind of mouse action is being 
		# performed and creates a "drag session" to handle it. The current drag session is stored
		# in the "drag_session" variable of the CircuitView. Drag sessions work in one of two ways:
		# 
		# 1. If the drag session uses Cocoa's drag & drop mechanism, then it will internally call
		# 	dragImage:at:offset:event:pasteboard:source:slideBack:, which blocks during the drag.
		#	When the CircuitView recieves NSDraggingDestination messages, it forwards them to the
		#	drag session (under different names). When the CircuitView draws itself, it also calls 
		#	the draw() method of the drag session in case it has anything to contribute. When the 
		#	drag is over, drag_session is set to None again. All this happens within mouseDown_().
		#
		# 2. Alternatively, the drag session may work with ordinary mouse events. In this case, the
		#	CircuitView forwards mouseDragged: and mouseUp: messages to the drag session (again
		#	under different names). The drag session remains in the drag_session variable of the
		#	CircuitView until mouseUp: is recieved. Again, the CircuitView gives the drag session
		#	an opportunity to contribute to the view's appearance by calling draw() in drawRect:.
		#
		# The following are the types of mouse actions and the corresponding drag session classes:
		#
		# 1. Rubber-band selection (handled by SelectDragSession, method #2)
		# 2. Creating a new wire (handled by WireDragSession, method #1)
		# 3. Moving components, no split wires (handled by FreeDragSession, method #1)
		# 4. Moving components, at least one split wire (handled by DependentDragSession, method #1)
		# 5. Manipulating controls (handled by ManipulateDragSession, method #2)
		#
	
		p = self.convertPoint_fromView_(event.locationInWindow(), None)
		
		def check_for_manipulation():
			# Check if we are clicking on a manipulatable spot on a component
			for obj in self.circuit.components:
				if point_in_rect(p, obj.bounds()):
					obj_p = untransform_point(obj.transform, p)
					if any([point_in_rect(obj_p, zone) for zone in obj.manipulation_zones]):
						obj.start_manipulation(obj_p)
						self.selection = set([obj])
						self.drag_session = ManipulateDragSession(self, obj)
						return True
			return False
		
		def clicking_on_selection():
			# Check if we are clicking near an existing member of the selection
			for obj in self.selection:
				if isinstance(obj, WireEnd):
					if point_distance(obj.location, p)<0.5:
						return True
				else:
					if point_in_rect(p, obj.bounds()):
						return True
			return False
		
		def clicking_near_object():
			# Check if we are clicking near an existing object
			for obj in self.circuit.components:
				if point_in_rect(p, obj.bounds()):
					return obj
			
			nearest_obj = None
			nearest_dist = 0.5 # Ignore any wires more than 8px away
			for wire in self.circuit.wires:
				for obj,other in [(wire.start, wire.end), (wire.end, wire.start)]:
					# Measure from a point that is not quite on the end of the wire so
					# that the user can pick one wire out of a group that all end at the
					# same point.
					handle = WireEnd.calc_handle_location(obj.location,
															 other.location, 0.1)
					d = point_distance(handle, p)
					if d < nearest_dist:
						nearest_obj = obj
						nearest_dist = d
			
			return nearest_obj
		
		if event.buttonNumber()==1 or event.modifierFlags()&NSCommandKeyMask:
			# We are dragging out a new wire
			self.drag_session = WireDragSession.alloc().initWithView_startLoc_(self, round_point(p))
			self.drag_session.run()
			self.drag_session = None
		
		elif check_for_manipulation():
			# check_for_manipulation() has done whatever is necessary
			pass
		
		elif clicking_on_selection():
			# We are dragging a group of already-selected objects
			self.mouseDownMove_(p)
			
		else:
			# We are either selecting a new object and dragging it, or making a rubber-band
			# selection.
			
			# Clear the current selection if the Command key is not held down
			if not event.modifierFlags() & NSCommandKeyMask:
				self.selection = set()
				
				# Redraw because selection is gone
				self.setNeedsDisplay_(True)
			
			# Figure out which object, if any, we are clicking on
			nearest_obj = clicking_near_object()
			
			if nearest_obj:
				# We have just selected an object and are now dragging it
				self.selection = self.selection.union(set([nearest_obj]))
				self.mouseDownMove_(p)
			
			else:
				# We are performing a rubber-band selection
				self.drag_session = SelectDragSession(self, p)
	
	def mouseDownMove_(self, p):
		# Determine whether there is at least one wire which is partially in the selection and
		# partially out of it. If there is, then the drag is considered "dependent", which means
		# that it can only end in the same view in which it started. This is to prevent people from
		# creating a wire between two different windows.
		dependent = False
		for part in self.selection:
			if isinstance(part, WireEnd) and part.other() not in self.selection:
				dependent = True
				break
		
		if dependent:
			self.drag_session = \
				DependentDragSession.alloc().initWithView_parts_(self, self.selection)
			self.drag_session.run() # Blocks until drag is complete
			self.drag_session = None
		
		else:
			self.drag_session = FreeDragSession.new_as_source(self, objs_of_parts(self.selection))
			self.drag_session.run() # Blocks until drag is complete
			self.drag_session = None
	
	@catch_exceptions
	def mouseDragged_(self, event):
		if self.drag_session:
			p = self.convertPoint_fromView_(event.locationInWindow(), None)
			self.drag_session.mouse_dragged(p)
	
	@catch_exceptions
	def mouseUp_(self, event):
		if self.drag_session:
			p = self.convertPoint_fromView_(event.locationInWindow(), None)
			self.drag_session.mouse_up(p)
			self.drag_session = None
		
	# Drag-destination role
	
	@catch_exceptions
	def draggingEntered_(self, sender):
	
		if self.drag_session:
			return self.drag_session.dragging_entered(sender)
		
		else:
			dataType = sender.draggingPasteboard().availableTypeFromArray_(["CircuitComponents"])
			
			if dataType == "CircuitComponents":
				# Here we are receiving a drag from another window. We create a temporary 
				# FreeDragSession which mirrors the FreeDragSession in the drag source, and set
				# drag_session_temporary to True so that it is discarded after the drag is over.
				#
				# This is slightly unusual because we are creating a drag session object even though 
				# the user did not click in this view.
				self.drag_session = FreeDragSession.new_as_dest(self, sender)
				self.drag_session_temporary = True
				return self.drag_session.dragging_entered(sender)
			
			else:
				return NSDragOperationNone
	
	@catch_exceptions
	def draggingUpdated_(self, sender):
		if self.drag_session:
			return self.drag_session.dragging_updated(sender)
		else:
			return NSDragOperationNone
	
	@catch_exceptions
	def draggingExited_(self, sender):
	
		if self.drag_session:
			self.drag_session.dragging_exited(sender)
			
			if self.drag_session_temporary:
				# Drag session was a temp created for receiving a drag.
				self.drag_session = None
				self.drag_session_temporary = False
	
	@catch_exceptions
	def prepareForDragOperation_(self, sender):
		return self.drag_session.prepare_for_drag_operation(sender)
	
	@catch_exceptions
	def performDragOperation_(self, sender):
		return self.drag_session.perform_drag(sender)
	
	@catch_exceptions
	def concludeDragOperation_(self, sender):
		self.drag_session.conclude_drag(sender)
		
		if self.drag_session_temporary:
			# Drag session was a temp created for receiving a drag.
			self.drag_session = None
			self.drag_session_temporary = False
